!******************************************************************************
!****m* /residue
! NAME
! module residue
!
! PURPOSE
! Construct target residue in an elementary particle - nucleus collision using
! particle-hole excitations.
! NOTES
! * Works both for non-relativistic Skyrme-mode and RMF-mode.
!******************************************************************************
module residue

  implicit none
  
  private

  !****************************************************************************
  !****g* residue/DetermineResidue
  ! PURPOSE
  ! If .true., then the determination of the target residue properties for every event and their output 
  ! in file TargetResidue.dat at the end of time evolution will be done
  ! SOURCE
  !
  logical, save ::  DetermineResidue = .false.
  !****************************************************************************

  
  !****************************************************************************
  !****t* residue/remnant
  ! NAME
  ! type remnant
  ! PURPOSE
  ! Stores properties of nuclear target remnant.
  !
  type remnant
     integer              :: mass=0      ! mass number   
     integer              :: charge=0    ! charge number
     real                 :: ExEnergy=0. ! excitation energy (GeV)
     real, dimension(1:3) :: momentum=0. ! momentum (GeV/c)
  end type remnant

      
  !****************************************************************************
  !****g* residue/TargetResidue
  ! SOURCE
  !
  type(remnant), allocatable, dimension(:,:), TARGET, SAVE :: TargetResidue
  ! PURPOSE
  ! Stores mass number, charge number, excitation energy and momentum of
  ! all target residues.
  !****************************************************************************

  type(remnant), pointer :: res
  
  public :: InitResidue,Add_ph,OutputResidue

contains

  !****************************************************************************
  !****s* residue/InitResidue
  ! NAME
  ! subroutine InitResidue(numEnsembles,numParticles,A,Z)
  ! PURPOSE
  ! * Initialize target residue.
  ! INPUTS
  ! *   integer, intent(in) :: numEnsembles,numParticles
  ! *   integer, intent(in) :: A,Z    ! target nucleus mass and charge numbers  
  ! NOTE
  ! Intention is to use this routine in electron- and photon-induced reactions, where the
  ! perturbative event is initialized on every nucleon of every ensemble. Thus there are two
  ! indices identifying the residue for respective perturbative event.
  !****************************************************************************
  subroutine InitResidue(numEnsembles,numParticles,A,Z)

      use output, only: Write_ReadingInput
    
      integer, intent(in) :: numEnsembles,numParticles
      integer, intent(in) :: A,Z    ! target nucleus mass and charge numbers

      integer :: ios
      
      !****n* residue/residue_Input
      ! NAME
      ! NAMELIST /residue_Input/
      ! PURPOSE
      ! Includes the switches:
      ! * DetermineResidue
      !**************************************************************************
      NAMELIST /residue_Input/ DetermineResidue

      call Write_ReadingInput('residue_Input',0)
      rewind(5)
      read(5,nml=residue_Input,iostat=ios)
      call Write_ReadingInput('residue_Input',0,ios)

      if(.not.DetermineResidue) return
 
      allocate(TargetResidue(1:numEnsembles,1:numParticles))
   
      TargetResidue%mass=A
      TargetResidue%charge=Z
      TargetResidue%ExEnergy=0.
      TargetResidue%momentum(1)=0.
      TargetResidue%momentum(2)=0.
      TargetResidue%momentum(3)=0.
      

  end subroutine InitResidue

    
  !****************************************************************************
  !****s* residue/Add_ph
  ! NAME
  ! subroutine Add_ph(firstEvent,teilchen)
  ! PURPOSE
  ! * Add a particle-hole excitation to the residue.
  ! INPUTS
  ! *   integer, intent(in) ::  firstEvent        ! Index of colliding particle
  ! *
  ! *   type(particle), intent(in) :: teilchen    ! Struck nucleon
  !****************************************************************************
  subroutine Add_ph(firstEvent,teilchen)

    use particleDefinition
    use densitymodule, only : FermiMomAt, energyDeterminationRMF
    use energyCalc, only : energyDetermination
    use RMF, only: getRMF_flag, g_rho
    use baryonPotentialModule, only: getsymmetryPotFlag_baryon
    
    integer, intent(in) :: firstEvent         ! Index of colliding particle
    type(particle), intent(in) :: teilchen    ! Struck nucleon

    type(particle) :: Teilchen_Fermi
    real :: pF
    integer :: iens,inuc

    if(.not.DetermineResidue) return

    iens=min(size(TargetResidue,dim=1),firstEvent/1000)
    inuc=min(size(TargetResidue,dim=2),firstEvent-iens*1000)
        
    res => TargetResidue(iens,inuc)

    res%mass=res%mass-1

    res%charge=res%charge-teilchen%charge

    if (getRMF_flag()) then
       
       if (g_rho/=0.) then
          pF=FermiMomAt(teilchen%position,teilchen%charge)
       else
          pF=FermiMomAt(teilchen%position)
       end if

    else
          
       if (getsymmetryPotFlag_baryon()) then
           pF=FermiMomAt(teilchen%position,teilchen%charge)
       else
           pF=FermiMomAt(teilchen%position)
       end if

    end if
    
    Teilchen_Fermi=teilchen
    Teilchen_Fermi%momentum(1:3)=(/pF,0.,0./)
    
    if(getRMF_flag()) then
       call energyDeterminationRMF(Teilchen_Fermi)
    else   
       call energyDetermination(Teilchen_Fermi)
    end if

    res%ExEnergy=res%ExEnergy+Teilchen_Fermi%momentum(0)-teilchen%momentum(0)

!    res%ExEnergy=res%ExEnergy+sqrt(pf**2+0.938**2)&
!         &-sqrt(teilchen%momentum(1)**2+teilchen%momentum(2)**2+teilchen%momentum(3)**2+0.938**2)

    res%momentum(1:3)=res%momentum(1:3)-teilchen%momentum(1:3)

  end subroutine Add_ph
  
    
  !****************************************************************************
  !****s* residue/OutputResidue
  ! NAME
  ! subroutine OutputResidue
  ! PURPOSE
  ! * Output of all residues to the file TargetResidue.dat
  !****************************************************************************
  subroutine OutputResidue

    integer :: iens,inuc

    if(.not.DetermineResidue) return
    
    open(1,file='TargetResidue.dat',position='append')
!    write(1,*)'#iens: inuc:  A:     Z:    E^* (GeV):       p(1:3) (GeV/c):'
    do iens=1,size(TargetResidue,dim=1)
       do inuc=1,size(TargetResidue,dim=2)
          res => TargetResidue(iens,inuc)
          write(1,'(1x,i5,3x,i3,4x,i3,3x,i3,3x,e14.7,3x,3(e14.7,1x))') &
               &iens,inuc,res%mass,res%charge,res%ExEnergy,res%momentum(1:3)
       end do
    end do
    close(1)

    deallocate(TargetResidue)
    
  end subroutine OutputResidue

end module residue

       
